// 
// Copyright (c) 2010 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Net.Pop3.Protocol.Client;
using Smdn.Net.Pop3.Client.Transaction;

namespace Smdn.Net.Pop3.Client.Session {
  partial class PopSession {
    /*
     * methods for processing transaction
     */
    private PopCommandResult ProcessTransaction(IPopTransaction t)
    {
      PreProcessTransaction(t, handlesIncapableAsException);

      if (transactionTimeout == Timeout.Infinite) {
        ProcessTransactionInternal(t);
      }
      else {
        if (!Task.Run(() => ProcessTransactionInternal(t)).Wait(transactionTimeout)) {
          CloseConnection();
          throw new TimeoutException(string.Format("transaction timeout ({0})", t.GetType().FullName));
        }
      }

      return PostProcessTransaction(t);
    }

    private void ProcessTransactionInternal(IPopTransaction t)
    {
      lock (transactionLockObject) {
        Trace.LogRequest(t);

        t.Process();

        Trace.LogResponse(t);
      }
    }

    private void PreProcessTransaction(IPopTransaction t, bool exceptionIfIncapable)
    {
      if (IsTransactionProceeding)
        throw new InvalidOperationException("another transaction proceesing");

      // check capability
      if (!exceptionIfIncapable)
        return;

      CheckServerCapability(t as IPopExtension);

      /*
      foreach (var arg in t.Request.Arguments.Values) {
        if (arg is IPopExtension)
          CheckServerCapability(arg as IPopExtension);
      }
      */
    }

    private PopCommandResult PostProcessTransaction(IPopTransaction t)
    {
      switch (t.Result.Code) {
        /*
         * disconnect without processing responses
         */
        case PopCommandResultCode.InternalError:
          CloseConnection();
          throw new PopException(t.Result.Description, t.Result.Exception);

        case PopCommandResultCode.SocketTimeout:
          CloseConnection();
          throw new TimeoutException(string.Format("socket timeout in {0}", t.GetType().FullName), t.Result.Exception);

        case PopCommandResultCode.ConnectionError:
        case PopCommandResultCode.UpgradeError:
          CloseConnection();
          throw t.Result.Exception;
      }

      return t.Result;
    }
  }
}
